from mock import patch, create_autospec, sentinel, Mock, PropertyMock, MagicMock
import pytest
from pymongo.errors import AutoReconnect, OperationFailure, DuplicateKeyError, ServerSelectionTimeoutError
from pymongo.read_preferences import ReadPreference

from arctic import decorators
from arctic.decorators import mongo_retry, _get_host
from pymongo.collection import Collection
from arctic.hooks import register_log_exception_hook


def test_mongo_retry():
    retries = [2]
    self = MagicMock()
    self._arctic_lib.arctic.mongo_host = sentinel.host
    self._collection.database.client.nodes = set([('a', 12)])
    self._arctic_lib.get_name.return_value = sentinel.lib_name
    with patch('arctic.decorators._handle_error', autospec=True) as he:
        @mongo_retry
        def foo(self):
            if retries[0] == 2:
                retries[0] -= 1
                raise OperationFailure('error')
            elif retries[0] == 1:
                retries[0] -= 1
                raise AutoReconnect('error')
            return "success"
        foo(self)
    assert he.call_count == 2
    assert isinstance(he.call_args_list[0][0][1], OperationFailure)
    assert he.call_args_list[0][0][2] == 1
    assert he.call_args_list[0][1] == {'mnodes': ['a:12'],
                                       'mhost': 'sentinel.host',
                                       'l': sentinel.lib_name}
    assert isinstance(he.call_args_list[1][0][1], AutoReconnect)
    assert he.call_args_list[1][0][2] == 2


def test_mongo_retry_hook_changes():
    retries = [2]
    self = MagicMock()
    hook1 = Mock()
    register_log_exception_hook(hook1)
    hook2 = Mock()

    @mongo_retry
    def foo(self):
        if retries[0] == 2:
            retries[0] -= 1
            raise OperationFailure('error')
        elif retries[0] == 1:
            register_log_exception_hook(hook2)
            retries[0] -= 1
            raise AutoReconnect('error')
        return "success"
    foo(self)

    assert hook1.call_count == 1
    assert hook2.call_count == 1


def test_mongo_retry_fails():
    error = OperationFailure('error')
    retries = [16]
    with patch('arctic.decorators._log_exception', autospec=True) as le:
        @mongo_retry
        def foo():
            if retries[0]:
                retries[0] -= 1
                raise error
            return "success"
        with pytest.raises(OperationFailure):
            foo()
    assert le.call_count == 15
    assert le.call_args[0][0] == 'foo'
    assert le.call_args[0][1] == error


def test_retry_nested():
    error = OperationFailure('error')
    with patch('arctic.decorators._log_exception', autospec=True) as le:
        @mongo_retry
        def foo():
            @mongo_retry
            def bar():
                raise error
            try:
                bar()
            except:
                raise error
        with pytest.raises(OperationFailure):
            foo()
    assert le.call_count == 15
    assert le.call_args[0][0] == 'bar'
    assert le.call_args[0][1] == error


def test_all_other_exceptions_logged():
    with patch('arctic.decorators._log_exception', autospec=True) as le:
        def foo():
            raise Exception("Unexpected Error")
        foo.__module__ = 'arctic.foo'
        foo = mongo_retry(foo)
        with pytest.raises(Exception) as e:
            foo()
    assert "Unexpected Error" in str(e)
    assert le.call_count == 1
    assert le.call_args[0][0] == "foo"


def test_other_exceptions_not_logged_outside_of_arctic():
    with patch('arctic.decorators._log_exception', autospec=True) as le:
        @mongo_retry
        def foo():
            raise Exception("Unexpected Error")
        with pytest.raises(Exception) as e:
            foo()
    assert "Unexpected Error" in str(e)
    assert le.call_count == 0


@pytest.mark.xfail(reason="CS-8393 Mongo server reports auth failure when servers flip")
def test_auth_failure_no_retry():
    error = OperationFailure('unauthorized for db:arctic_jblackburn')
    with patch('arctic.decorators._log_exception', autospec=True) as le:
        @mongo_retry
        def foo():
            raise error
        with pytest.raises(OperationFailure) as e:
            foo()
    assert 'OperationFailure: unauthorized for db:arctic_jblackburn' in str(e)
    assert le.call_count == 1


def test_duplicate_key_failure_no_retry():
    error = DuplicateKeyError('duplicate key')
    with patch('arctic.decorators._log_exception', autospec=True) as le:
        @mongo_retry
        def foo():
            raise error
        with pytest.raises(OperationFailure) as e:
            foo()
    assert 'duplicate key' in str(e)
    assert le.call_count == 1


def test_ServerSelectionTimeoutError_no_retry():
    error = ServerSelectionTimeoutError('some error')
    with patch('arctic.decorators._log_exception', autospec=True) as le:
        @mongo_retry
        def foo():
            raise error
        with pytest.raises(ServerSelectionTimeoutError) as e:
            foo()
    assert 'some error' in str(e)
    assert le.call_count == 1


def test_get_host():
    store = Mock()
    store._arctic_lib.arctic.mongo_host = sentinel.host
    store._collection.database.client.nodes = set([('a', 12)])
    store._arctic_lib.get_name.return_value = sentinel.lib_name
    assert _get_host(store) == {'mhost': 'sentinel.host',
                                'mnodes': ['a:12'],
                                'l': sentinel.lib_name,
                               }


def test_get_host_list():
    store = Mock()
    store._arctic_lib.arctic.mongo_host = sentinel.host
    store._collection.database.client.nodes = set([('a', 12)])
    store._arctic_lib.get_name.return_value = sentinel.lib_name
    assert _get_host([store]) == {'mhost': 'sentinel.host',
                                 'mnodes': ['a:12'],
                                 'l': sentinel.lib_name,
                                 }


def test_get_host_not_a_vs():
    store = MagicMock()
    store._arctic_lib.get_name.side_effect = AttributeError("Hello")
    assert _get_host(store) == {}
    store._arctic_lib.get_name.side_effect = ValueError("Hello")
    assert _get_host(store) == {}



from __future__ import print_function

################################################
# Getting started
################################################

# Install Arctic
#    pip install git+https://github.com/manahl/arctic.git
# That's it(!)

# Run MongoDB - https://www.mongodb.org/downloads
#    $ mkdir /tmp/pydata-demo
#    $ mongod --dbpath /tmp/pydata-demo

from datetime import datetime
import time
import ystockquote

from arctic import Arctic
import collections
import pandas
import pandas as pd
import pprint


################################################
# Loading data
################################################


def get_stock_history(ticker, start_date, end_date):
    data = ystockquote.get_historical_prices(ticker, start_date, end_date)
    df = pandas.DataFrame(collections.OrderedDict(sorted(data.items()))).T
    df = df.convert_objects(convert_numeric=True)
    return df


################################################
# VersionStore: Storing and updating stock data
################################################



arctic = Arctic('localhost')

# Create a VersionStore library
arctic.delete_library('jblackburn.stocks')
arctic.initialize_library('jblackburn.stocks')
arctic.list_libraries()


stocks = arctic['jblackburn.stocks']

# get some prices
aapl = get_stock_history('aapl', '2015-01-01', '2015-02-01')
aapl

# store them in the library
stocks.write('aapl', aapl, metadata={'source': 'YAHOO'})
stocks.read('aapl').data['Adj Close'].plot()
stocks.read('aapl').metadata
stocks.read('aapl').version


# Append some more prices - imagine doing this once per period
aapl = get_stock_history('aapl', '2015-02-01', '2015-03-01')
stocks.append('aapl', aapl)
stocks.read('aapl').data


# Reading different versions of the symbol
stocks.list_symbols()
stocks.list_versions('aapl')

# Read the different versions separately
stocks.read('aapl', as_of=1).data.ix[-1]
stocks.read('aapl', as_of=2).data.ix[-1]


# And we can snapshot all items in the library
stocks.snapshot('snap')
stocks.read('aapl', as_of='snap').data.ix[-1]



#################################
# Dealing with lots of data
#################################


#NSYE library
lib = arctic['nyse']

def load_all_stock_history_NYSE():
    # Data downloaded from BBG Open Symbology:
    # 
    nyse = pd.read_csv('/users/is/jblackburn/git/arctic/howtos/nyse.csv')
    stocks = [x.split('/')[0] for x in nyse['Ticker']]
    print(len(stocks), " symbols")
    for i, stock in enumerate(stocks):
        try:
            now = datetime.now()
            data = get_stock_history('aapl', '1980-01-01', '2015-07-07')
            lib.write(stock, data)
            print("loaded data for: ", stock, datetime.now() - now)
        except Exception as e:
            print("Failed for ", stock, str(e))


# load_all_stock_history_NYSE()
print(len(lib.list_symbols()), " NYSE symbols loaded")


def read_all_data_from_lib(lib):
    start = time.time()
    rows_read = 0
    for s in lib.list_symbols():
        rows_read += len(lib.read(s).data)
    print("Symbols: %s Rows: %s  Time: %s  Rows/s: %s" % (len(lib.list_symbols()),
                                                          rows_read,
                                                          (time.time() - start),
                                                          rows_read / (time.time() - start)))


read_all_data_from_lib(lib)
# Symbols: 1315 Rows: 11460225   Rows/s: 2,209,909



from __future__ import print_function
import random
import lz4
import string
import pytest
import six
from datetime import datetime as dt

import arctic._compress as c


@pytest.mark.parametrize("n, length", [(300, 5e4),  # micro TS
                                       (5, 2e6),  # Futures TS
                                       (10, 2e6),  # Futures TS
                                       (100, 2e6),  # Large TS
                                       (250, 2e6)])  # Even Bigger TS
def test_performance_sequential(n, length):
    _str = random_string(length)
    _strarr = [_str for _ in range(n)]
    now = dt.now()
    [c.decompress(y) for y in [c.compressHC(x) for x in _strarr]]
    clz4_time = (dt.now() - now).total_seconds()
    now = dt.now()
    c.decompressarr(c.compressarrHC(_strarr))
    clz4_time_p = (dt.now() - now).total_seconds()
    now = dt.now()
    [lz4.decompress(y) for y in [lz4.compressHC(x) for x in _strarr]]
    lz4_time = (dt.now() - now).total_seconds()
    print()
    print("LZ4 Test %sx len:%s" % (n, length))
    print("    Cython LZ4 %s s" % clz4_time)
    print("    Cython LZ4 Parallel %s s" % clz4_time_p)
    print("    LZ4 %s s" % lz4_time)


def random_string(N):
    _str = ''.join(random.choice(list(string.printable) + ['hello', 'world', 'hellworld', 'Hello', 'w0rld']) for _ in six.moves.xrange(int(N)))
    return _str.encode('ascii')


def test_exceptions():
    data = c.compress(b'1010101010100000000000000000000000000000000000000000000000000000000011111111111111111111111111111')
    data = data[0:16]
    with pytest.raises(Exception) as e:
        c.decompress(data)
    assert("decompressing" in str(e))

    data = c.compress(b'1010101010100000000000000000000000000000000000000000000000000000000011111111111111111111111111111')
    data = [data[0:16] for x in (1, 2, 3)]
    with pytest.raises(Exception) as e:
        c.decompressarr(data)
    assert("decompressing" in str(e))
